/*
 *
 *
 *    envdet.h
 *
 *    Unipolar envelope detector.
 *
 *    Copyright (C) 2016 by Matt Roberts.
 *    License: GNU GPL3 (www.gnu.org)
 *
 *
 */


#ifndef __KK5JY_ENVDET_H
#define __KK5JY_ENVDET_H

#include <stdint.h>
#include <math.h>

#include "ctrthr.h"
#include "nlimits.h"
#include "minmax.h"

namespace KK5JY {
	namespace DSP {
		//
		//  EnvelopeDetector
		//
		template <typename sample_t, typename ctr_t = int8_t>
		class EnvelopeDetector {
			private:
				// build the min/max tracker
				KK5JY::DSP::MinMaxTrack<sample_t> minmax;

				// build the output counter
				KK5JY::DSP::CounterThreshold<ctr_t> ctrthr;

				// running magnitude envelope
				sample_t mmin;
				sample_t mmax;

				// current threshold
				sample_t threshold;

			public:
				EnvelopeDetector(size_t mm_win, size_t ctr_lim);

				bool run(sample_t sample);
		};


		//
		//  ctor
		//
		template <typename sample_t, typename ctr_t>
		inline EnvelopeDetector<sample_t, ctr_t>::EnvelopeDetector(size_t mm_win, size_t ctr_lim)
			: minmax(mm_win),
			  ctrthr(ctr_lim),
			  mmin(0),
			  mmax(KK5JY::DSP::norm_limits<sample_t>::maximum),
			  threshold(0) {
			// nop
		}


		//
		//  run
		//
		template <typename sample_t, typename ctr_t>
		inline bool EnvelopeDetector<sample_t, ctr_t>::run(sample_t value) {
			// min/max detection
			if (minmax.run(value)) {
				mmax = minmax.maximum();
				mmin = minmax.minimum();

				// minimum must be <= 0
				if (mmin > static_cast<sample_t>(0))
					mmin = static_cast<sample_t>(0);

				// compute new threshold
				threshold = (((mmax - mmin) >> 1) + mmin);
			}

			// run the output counter
			return ctrthr.run(value >= threshold ? 1 : -1);
		}


		//
		//  run
		//
		template <>
		inline bool EnvelopeDetector<double, int8_t>::run(double value) {
			// min/max detection
			if (minmax.run(value)) {
				mmax = minmax.maximum();
				mmin = minmax.minimum();

				// minimum must be <= 0
				if (mmin > static_cast<double>(0))
					mmin = static_cast<double>(0);

				// compute new threshold
				threshold = ((0.5 * (mmax - mmin)) + mmin);
			}

			// run the output counter
			return ctrthr.run(value >= threshold ? 1 : -1);
		}


		//
		//  run
		//
		template <>
		inline bool EnvelopeDetector<float, int8_t>::run(float value) {
			// min/max detection
			if (minmax.run(value)) {
				mmax = minmax.maximum();
				mmin = minmax.minimum();

				// minimum must be <= 0
				if (mmin > static_cast<float>(0))
					mmin = static_cast<float>(0);

				// compute new threshold
				threshold = ((0.5 * (mmax - mmin)) + mmin);
			}

			// run the output counter
			return ctrthr.run(value >= threshold ? 1 : -1);
		}
	}
}

#endif
